/*
 * list_file.C
 * Macro to list the content of ROOT TFiles and to save to ASCII CSV files all
 * the TNtuples that are found.
 *
 * Copyright (C) 2009-2015, Cristiano Lino Fontana
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University of Padua nor the
 *    names of its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CRISTIANO LINO FONTANA BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Changelog:
 * 2009/08/23 - First version
 * 2015/11/06 - Added the option to save ntuples to file
 * 2015/11/09 - Converted from standalone to ROOT macro
 *
 */

#include <iostream>
#include <fstream>

#include <string>

#include "Rtypes.h"
#include "TList.h"
#include "TObjArray.h"
#include "TFile.h"
#include "TKey.h"
#include "TDirectoryFile.h"
#include "TTree.h"
#include "TBranch.h"
#include "TNtuple.h"

void list_directory(TDirectoryFile *input_directory, TList *ntuples);

int list_file(const char *file_name, bool save_data = true)
{
        std::cout << "Opening file: " << file_name << std::endl;

        TList ntuples;

        TFile in_file(file_name, "READ");
        if (!in_file.IsOpen())
        {
            std::cerr << "ERROR, unable to open: " << file_name << std::endl;

            return EXIT_FAILURE;
        }

        TDirectoryFile *main_directory = reinterpret_cast<TDirectoryFile*>(&in_file);

        list_directory(main_directory, &ntuples);
    
        if (save_data)
        {
            TIter next(&ntuples);

            while (TNtuple *ntuple = reinterpret_cast<TNtuple*>(next()))
            {
                const Char_t *ntuple_name = ntuple->GetName();

                std::string out_file_name(file_name);

                const std::size_t found = out_file_name.rfind(".root");
                if (found != std::string::npos)
                {
                    out_file_name.erase(out_file_name.begin() + found, out_file_name.end());
                }

                out_file_name.append("_");
                out_file_name.append(ntuple_name);
                out_file_name.append(".csv");

                std::cout << "Saving TNtuple " << ntuple_name << " to: " << out_file_name << std::endl;

                std::ofstream out_file(out_file_name, std::ofstream::out);

                const TObjArray *branches = ntuple->GetListOfBranches();
                const Int_t n_branches = ntuple->GetNbranches();

                Float_t *data = new Float_t[n_branches];

                for (int i = 0; i < n_branches; i++)
                {
                    TBranch *branch = reinterpret_cast<TBranch*>(branches->At(i));

                    branch->SetAddress(data + i);
                }

                const Long64_t n_entries = ntuple->GetEntries();
                for (Long64_t j = 0; j < n_entries; j++)
                {
                    ntuple->GetEntry(j);

                    for (int i = 0; i < n_branches; i++)
                        out_file << data[i] << ';';

                    out_file << std::endl;
                }

                delete [] data;

                out_file.close();
            }
        }

        in_file.Close();

        return EXIT_SUCCESS;
}

void list_directory(TDirectoryFile *input_directory, TList *ntuples)
{
    input_directory->ls();

    const TList *list_of_keys = input_directory->GetListOfKeys();

    TIter next(list_of_keys);

    while (TKey *key = reinterpret_cast<TKey*>(next()))
    {
        const Char_t *name = key->GetName();

        TObject *object = input_directory->Get(name);

        const std::string class_name(const_cast<Char_t*>(object->ClassName()));

        if (class_name.compare("TDirectoryFile") == 0)
        {
            TDirectoryFile *directory = reinterpret_cast<TDirectoryFile*>(object);

            list_directory(directory, ntuples);
        }
        else if (class_name.compare("TTree") == 0)
        {
            TTree *tree = reinterpret_cast<TTree*>(object);

            tree->Print();
        }
        else if (class_name.compare("TNtuple") == 0)
        {
            TNtuple *ntuple = reinterpret_cast<TNtuple*>(object);

            ntuple->Print();

            ntuples->Add(ntuple);
        }
    }
}
